﻿// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved

using System;
using System.IO;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Graphics.Display;
using Windows.Media.Capture;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.System;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Media.Imaging;

namespace Microsoft.Samples.WebcamExperiences.CameraCaptureUICS
{
    partial class MainPage
    {
        public MainPage()
        {
            InitializeComponent();
            ScenarioList.SelectionChanged += new SelectionChangedEventHandler(ScenarioList_SelectionChanged);

            // Starting scenario is the first or based upon a previous selection
            ListBoxItem startingScenario = null;
            if (SuspensionManager.SessionState.ContainsKey("SelectedScenario"))
            {
                String selectedScenarioName = SuspensionManager.SessionState["SelectedScenario"] as string;
                startingScenario = this.FindName(selectedScenarioName) as ListBoxItem;
            }

            ScenarioList.SelectedItem = startingScenario != null ? startingScenario : Scenario1;

            App.Current.Suspending += new SuspendingEventHandler(App_Suspending);

            Scenario1InitAsync();
            Scenario2InitAsync();

            Windows.UI.ViewManagement.ApplicationView.GetForCurrentView().ViewStateChanged += new TypedEventHandler<Windows.UI.ViewManagement.ApplicationView, Windows.UI.ViewManagement.ApplicationViewStateChangedEventArgs>(MainPage_LayoutChanged);
            DisplayProperties.OrientationChanged += new DisplayPropertiesEventHandler(DisplayProperties_OrientationChanged);
        }

        void DisplayProperties_OrientationChanged(object sender)
        {
            if (DisplayProperties.CurrentOrientation == DisplayOrientations.Portrait || DisplayProperties.CurrentOrientation == DisplayOrientations.PortraitFlipped)
            {
                VisualStateManager.GoToState(this, "Portrait", false);
            }
        }

        void MainPage_LayoutChanged(Windows.UI.ViewManagement.ApplicationView sender, Windows.UI.ViewManagement.ApplicationViewStateChangedEventArgs args)
        {
            switch (args.ViewState)
            {
                case Windows.UI.ViewManagement.ApplicationViewState.Filled:
                    VisualStateManager.GoToState(this, "Fill", false);
                    break;
                case Windows.UI.ViewManagement.ApplicationViewState.FullScreenLandscape:
                    VisualStateManager.GoToState(this, "Full", false);
                    break;
                case Windows.UI.ViewManagement.ApplicationViewState.Snapped:
                    VisualStateManager.GoToState(this, "Snapped", false);
                    break;
                case Windows.UI.ViewManagement.ApplicationViewState.FullScreenPortrait:
                    VisualStateManager.GoToState(this, "Portrait", false);
                    break;
                default:
                    break;
            }
        }

        #region Scenario Specific Code
        private string photoFile = "";
        private string videoFile = "";

        async Task Scenario1InitAsync()
        {
            Scenario1Button.Click += new RoutedEventHandler(Scenario1Button_Click);
            Scenario1ResetButton.Click += new RoutedEventHandler(Scenario1ResetButton_Click);

            // Resume user's state
            if (SuspensionManager.SessionState.ContainsKey("AppPhotoFile"))
            {
                string filePath = SuspensionManager.SessionState["AppPhotoFile"] as string;
                if (filePath != "")
                {
                    photoFile = filePath;
                    StorageFile file = await StorageFile.GetFileFromPathAsync(filePath);
                    BitmapImage bitmapImage = new BitmapImage();
                    using(IRandomAccessStream fileStream = await file.OpenAsync(Windows.Storage.FileAccessMode.Read))
                    {
                        bitmapImage.SetSource(fileStream);
                    }
                    Scenario1Image.Source = bitmapImage;
                    Scenario1OutputText.Text = "";
                    Scenario1ResetButton.Visibility = Visibility.Visible;
                }
            }
        }

        async void Scenario1Button_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                CameraCaptureUI dialog = new CameraCaptureUI();
                Size aspectRatio = new Size(16, 9);
                dialog.PhotoSettings.CroppedAspectRatio = aspectRatio;

                StorageFile file = await dialog.CaptureFileAsync(CameraCaptureUIMode.Photo);
                if (file != null)
                {
                    Scenario1OutputText.Text = "";
                    BitmapImage bitmapImage = new BitmapImage();
                    using(IRandomAccessStream fileStream = await file.OpenAsync(Windows.Storage.FileAccessMode.Read))
                    {
                        bitmapImage.SetSource(fileStream);
                    }
                    Scenario1Image.Source = bitmapImage;
                    Scenario1ResetButton.Visibility = Visibility.Visible;
                    photoFile = file.Path;
                }
                else
                {
                    Scenario1OutputText.Text = "No photo captured.";
                }
            }
            catch (Exception ex)
            {
                Scenario1OutputText.Text = ex.Message;
            }
        }

        void Scenario1ResetButton_Click(object sender, RoutedEventArgs e)
        {
            photoFile = "";
            Scenario1Reset();
            Scenario1Image.Source = new BitmapImage(new Uri(this.BaseUri, "images/placeholder-sdk.png"));
        }

        void Scenario1Reset()
        {
            if (photoFile == "")
            {
                Scenario1OutputText.Text = "The photo will be shown here.";
                Scenario1ResetButton.Visibility = Visibility.Collapsed;
            }
        }

        async Task Scenario2InitAsync()
        {
            Scenario2Button.Click += new RoutedEventHandler(Scenario2Button_Click);
            Scenario2ResetButton.Click += new RoutedEventHandler(Scenario2ResetButton_Click);

            // Resume user's state
            if (SuspensionManager.SessionState.ContainsKey("AppVideoFile"))
            {
                string filePath = SuspensionManager.SessionState["AppVideoFile"] as string;
                if (filePath != "")
                {
                    videoFile = filePath;
                    StorageFile file = await StorageFile.GetFileFromPathAsync(filePath);
                    IRandomAccessStream fileStream = await file.OpenAsync(Windows.Storage.FileAccessMode.Read);
                    Scenario2Video.SetSource(fileStream, "video/mp4");
                    Scenario2OutputText.Text = "";
                    Scenario2ResetButton.Visibility = Visibility.Visible;
                }
            }
        }

        async void Scenario2Button_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                CameraCaptureUI dialog = new CameraCaptureUI();
                dialog.VideoSettings.Format = CameraCaptureUIVideoFormat.Mp4;

                StorageFile file = await dialog.CaptureFileAsync(CameraCaptureUIMode.Video);
                if (file != null)
                {
                    Scenario2OutputText.Text = "";
                    IRandomAccessStream fileStream = await file.OpenAsync(Windows.Storage.FileAccessMode.Read);
                    Scenario2Video.SetSource(fileStream, "video/mp4");
                    Scenario2ResetButton.Visibility = Visibility.Visible;
                    videoFile = file.Path;
                }
                else
                {
                    Scenario2OutputText.Text = "No video captured.";
                }
            }
            catch (Exception ex)
            {
                Scenario2OutputText.Text = ex.Message;
            }
        }

        void Scenario2ResetButton_Click(object sender, RoutedEventArgs e)
        {
            videoFile = "";
            Scenario2Reset();
            Scenario2Video.Source = null;
        }

        void Scenario2Reset()
        {
            if (videoFile == "")
            {
                Scenario2OutputText.Text = "The video will be shown here.";
                Scenario2ResetButton.Visibility = Visibility.Collapsed;
            }
        }

        void App_Suspending(object sender, Windows.ApplicationModel.SuspendingEventArgs e)
        {
            // Save user's state
            SuspensionManager.SessionState["AppPhotoFile"] = photoFile;
            SuspensionManager.SessionState["AppVideoFile"] = videoFile;
        }
        #endregion

        void ScenarioList_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            ResetAll();
            if (ScenarioList.SelectedItem == Scenario1)
            {
                Scenario1Input.Visibility = Visibility.Visible;
                Scenario1Output.Visibility = Visibility.Visible;
            }
            else if (ScenarioList.SelectedItem == Scenario2)
            {
                Scenario2Input.Visibility = Visibility.Visible;
                Scenario2Output.Visibility = Visibility.Visible;
            }

            if (ScenarioList.SelectedItem != null)
            {
                ListBoxItem selectedListBoxItem = ScenarioList.SelectedItem as ListBoxItem;
                SuspensionManager.SessionState["SelectedScenario"] = selectedListBoxItem.Name;
            }
        }

        public void ResetAll()
        {
            Scenario1Input.Visibility = Visibility.Collapsed;
            Scenario1Output.Visibility = Visibility.Collapsed;
            Scenario1Reset();
            Scenario2Input.Visibility = Visibility.Collapsed;
            Scenario2Output.Visibility = Visibility.Collapsed;
            Scenario2Reset();
        }

        async void Footer_Click(object sender, RoutedEventArgs e)
        {
            await Windows.System.Launcher.LaunchUriAsync(new Uri(((HyperlinkButton)sender).Tag.ToString()));
        }
    }
}
